/* Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * This is the file which integrates between 'gotest' and 'gocheck' support for
 * creating suites of tests, etc.
 *
 */

package dirstate
// This is the whitebox tests for the State class

import (
	. "launchpad.net/gocheck"
)

// This is the basic Suite type that all functions integrate around.
type WhiteboxSuite struct{}

var _ = Suite(&WhiteboxSuite{})


func (s *WhiteboxSuite) TestNew(c *C) {
	state := New()
	c.Assert(state.headerState, Equals, NOT_IN_MEMORY)
	c.Assert(state.dirblockState, Equals, NOT_IN_MEMORY)
	c.Assert(state.dirblocks, Equals, make([]block, 0))
}

func (s *WhiteboxSuite) TestCreateRoot(c *C) {
	state := New()
	state.CreateRoot(ROOT_ID)
	c.Assert(state.headerState, Equals, IN_MEMORY_MODIFIED)
	c.Assert(state.dirblockState, Equals, IN_MEMORY_MODIFIED)
	c.Assert(len(state.dirblocks), Equals, 2)
	root_block := state.dirblocks[0]
	c.Assert(root_block.dirname, Equals, "")
	c.Assert(len(root_block.entries), Equals, 1)
	root_entry := root_block.entries[0]
	c.Assert(root_entry.key.dirname, Equals, "")
	c.Assert(root_entry.key.basename, Equals, "")
	c.Assert(root_entry.key.fileID, Equals, ROOT_ID)
	// Why is 'd' an int, and not a byte? Equals says we were expecting an int,
	// but found a uint, and though both are "100", they don't match.
	c.Assert(root_entry.work.details.minikind, Equals, byte('d'))
}


func assertNext(c *C, p *entryParser, expected string) {
	content, err := p.getNext()
	c.Assert(err, IsNil)
	c.Assert(content, Equals, []byte(expected))
}

func assertNextFail(c *C, p *entryParser, expected_err string) {
	content, err := p.getNext()
	c.Assert(err, NotNil)
	c.Assert(content, IsNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetNext(c *C) {
	text := []byte("some\x00text\x00")
	parser := newEntryParser(text, 0)
	assertNext(c, parser, "some")
	assertNext(c, parser, "text")
	assertNextFail(c, parser, "Could not find trailing null byte")
}

func (s *WhiteboxSuite) TestEntryParserGetKey(c *C) {
	text := []byte("dir\x00name\x00file_id\x00")
	parser := newEntryParser(text, 0)
	key := entryKey{}
	err := parser.getKey(&key)
	c.Assert(err, IsNil)
	c.Assert(key.dirname, Equals, "dir")
	c.Assert(key.basename, Equals, "name")
	c.Assert(key.fileID, Equals, "file_id")
}

func assertParserGetKeyFails(c *C, text string, expected_err string) {
	text_b := []byte(text)
	parser := newEntryParser(text_b, 0)
	key := entryKey{}
	err := parser.getKey(&key)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetKeyFailures(c *C) {
	assertParserGetKeyFails(c, "dir\x00name\x00file_id",
		"Unable to find fileID for key: Could not find trailing null byte")
	assertParserGetKeyFails(c, "dir\x00name\x00",
		"Unable to find fileID for key: Could not find trailing null byte")
	assertParserGetKeyFails(c, "dir\x00name",
		"Unable to find basename for key: Could not find trailing null byte")
	assertParserGetKeyFails(c, "dir\x00",
		"Unable to find basename for key: Could not find trailing null byte")
	assertParserGetKeyFails(c, "dir",
		"Unable to find dirname for key: Could not find trailing null byte")
	assertParserGetKeyFails(c, "",
		"Unable to find dirname for key: Could not find trailing null byte")
}

func (s *WhiteboxSuite) TestEntryParserGetDetails(c *C) {
	text := []byte("d\x00fingerprint\x000\x00n\x00")
	parser := newEntryParser(text, 0)
	d := details{}
	err := parser.getDetails(&d)
	c.Assert(err, IsNil)
	c.Assert(d.minikind, Equals, byte('d'))
	c.Assert(d.fingerprint, Equals, "fingerprint")
	c.Assert(d.size, Equals, uint64(0))
	c.Assert(d.executable, Equals, false)
}

func assertParserGetDetailsFails(c *C, text string, expected_err string) {
	text_b := []byte(text)
	parser := newEntryParser(text_b, 0)
	d := details{}
	err := parser.getDetails(&d)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetDetailsFailures(c *C) {
	assertParserGetDetailsFails(c, "",
		"minikind missing: Could not find trailing null byte")
	assertParserGetDetailsFails(c, "long\x00fingerprint\x000\x00n\x00",
		"minikind must be a single byte")
	assertParserGetDetailsFails(c, "\x00fingerprint\x000\x00n\x00",
		"minikind must be a single byte")
	assertParserGetDetailsFails(c, "d\x00fingerprint",
		"fingerprint missing: Could not find trailing null byte")
	assertParserGetDetailsFails(c, "d\x00fingerprint\x000",
		"size missing: Could not find trailing null byte")
	assertParserGetDetailsFails(c, "d\x00fingerprint\x00abcd\x00",
		"invalid size: parsing \"abcd\": invalid argument")
	assertParserGetDetailsFails(c,
		"d\x00fingerprint\x0018446744073709551616\x00",
		"invalid size: parsing \"18446744073709551616\":"+
			" numerical result out of range")
	assertParserGetDetailsFails(c, "d\x00fingerprint\x000\x00",
		"executable missing: Could not find trailing null byte")
	assertParserGetDetailsFails(c, "d\x00fingerprint\x000\x00yes\x00",
		"executable must be a single byte long")
	assertParserGetDetailsFails(c, "d\x00fingerprint\x000\x00x\x00",
		"executable must be either 'n' or 'y'")
}

func (s *WhiteboxSuite) TestEntryParserGetWorkingDetails(c *C) {
	text := []byte("d\x00fingerprint\x000\x00n\x001234567890abcdef\x00")
	parser := newEntryParser(text, 0)
	wd := workingDetails{}
	err := parser.getWorkingDetails(&wd)
	c.Assert(err, IsNil)
	c.Assert(wd.minikind, Equals, byte('d'))
	c.Assert(wd.fingerprint, Equals, "fingerprint")
	c.Assert(wd.size, Equals, uint64(0))
	c.Assert(wd.executable, Equals, false)
	c.Assert(wd.packedHash, Equals, "1234567890abcdef")
}

// TODO: How would I write a helper function that can look up the function
//       to call in a dynamic fashion? I guess it also needs a dynamic type
//       argument, so maybe it isn't worth the overhead.
func assertParserGetWorkingDetailsFails(c *C, text string, expected_err string) {
	text_b := []byte(text)
	parser := newEntryParser(text_b, 0)
	wd := workingDetails{}
	err := parser.getWorkingDetails(&wd)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetWorkingDetailsFailures(c *C) {
	base := "d\x00fingerprint\x000\x00n\x00"
	assertParserGetWorkingDetailsFails(c, "",
		"minikind missing: Could not find trailing null byte")
	assertParserGetWorkingDetailsFails(c, "d\x00",
		"fingerprint missing: Could not find trailing null byte")
	assertParserGetWorkingDetailsFails(c, base,
		"packedHash missing: Could not find trailing null byte")
	assertParserGetWorkingDetailsFails(c, base+"deadbeef",
		"packedHash missing: Could not find trailing null byte")
}

func (s *WhiteboxSuite) TestEntryParserGetParentDetails(c *C) {
	text := []byte("d\x00fingerprint\x000\x00n\x00revision-id-abcdefgh\x00")
	parser := newEntryParser(text, 0)
	pd := parentDetails{}
	err := parser.getParentDetails(&pd)
	c.Assert(err, IsNil)
	c.Assert(pd.minikind, Equals, byte('d'))
	c.Assert(pd.fingerprint, Equals, "fingerprint")
	c.Assert(pd.size, Equals, uint64(0))
	c.Assert(pd.executable, Equals, false)
	c.Assert(pd.revisionID, Equals, "revision-id-abcdefgh")
}

// TODO: How would I write a helper function that can look up the function
//       to call in a dynamic fashion? I guess it also needs a dynamic type
//       argument, so maybe it isn't worth the overhead.
func assertParserGetParentDetailsFails(c *C, text string, expected_err string) {
	text_b := []byte(text)
	parser := newEntryParser(text_b, 0)
	pd := parentDetails{}
	err := parser.getParentDetails(&pd)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetParentDetailsFailures(c *C) {
	base := "d\x00fingerprint\x000\x00n\x00"
	assertParserGetParentDetailsFails(c, "",
		"minikind missing: Could not find trailing null byte")
	assertParserGetParentDetailsFails(c, "d\x00",
		"fingerprint missing: Could not find trailing null byte")
	assertParserGetParentDetailsFails(c, base,
		"revisionID missing: Could not find trailing null byte")
	assertParserGetParentDetailsFails(c, base+"deadbeef",
		"revisionID missing: Could not find trailing null byte")
}

func (s *WhiteboxSuite) TestEntryParserGetEntryNoParents(c *C) {
	text := []byte("dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"\n\x00")
	parser := newEntryParser(text, 0)
	e := entry{}
	err := parser.getEntry(&e)
	c.Assert(err, IsNil)
	c.Assert(e.key.dirname, Equals, "dir")
	c.Assert(e.key.basename, Equals, "name")
	c.Assert(e.key.fileID, Equals, "file_id")
	c.Assert(e.work.minikind, Equals, byte('f'))
	c.Assert(e.work.fingerprint, Equals, "fingerprint")
	c.Assert(e.work.size, Equals, uint64(1234))
	c.Assert(e.work.executable, Equals, true)
	c.Assert(e.work.packedHash, Equals, "abcd124befbde41392409")
	c.Assert(e.basis.minikind, Equals, byte(0))
	c.Assert(e.basis.fingerprint, Equals, "")
	c.Assert(e.basis.size, Equals, uint64(0))
	c.Assert(e.basis.executable, Equals, false)
	c.Assert(e.basis.revisionID, Equals, "")
	c.Assert(len(e.extra), Equals, 0)
}

func (s *WhiteboxSuite) TestEntryParserGetEntryWBasis(c *C) {
	text := []byte("dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"f\x00abcd124defbeef\x009999\x00y\x00rev-id\x00" +
		"\n\x00")
	parser := newEntryParser(text, 1)
	e := entry{}
	err := parser.getEntry(&e)
	c.Assert(err, IsNil)
	c.Assert(e.key.dirname, Equals, "dir")
	c.Assert(e.key.basename, Equals, "name")
	c.Assert(e.key.fileID, Equals, "file_id")
	c.Assert(e.work.minikind, Equals, byte('f'))
	c.Assert(e.work.fingerprint, Equals, "fingerprint")
	c.Assert(e.work.size, Equals, uint64(1234))
	c.Assert(e.work.executable, Equals, true)
	c.Assert(e.work.packedHash, Equals, "abcd124befbde41392409")
	c.Assert(e.basis.minikind, Equals, byte('f'))
	c.Assert(e.basis.size, Equals, uint64(9999))
	c.Assert(e.basis.executable, Equals, true)
	c.Assert(e.basis.revisionID, Equals, "rev-id")
	c.Assert(len(e.extra), Equals, 0)
}

func (s *WhiteboxSuite) TestEntryParserGetEntryWExtra(c *C) {
	text := []byte("dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"f\x00abcd124defbeef\x009999\x00y\x00rev-id\x00" +
		"d\x00defbeef\x008888\x00n\x00alt-rev-id\x00" +
		"\n\x00")
	parser := newEntryParser(text, 2)
	e := entry{}
	err := parser.getEntry(&e)
	c.Assert(err, IsNil)
	c.Assert(e.key.dirname, Equals, "dir")
	c.Assert(e.key.basename, Equals, "name")
	c.Assert(e.key.fileID, Equals, "file_id")
	c.Assert(e.work.minikind, Equals, byte('f'))
	c.Assert(e.work.fingerprint, Equals, "fingerprint")
	c.Assert(e.work.size, Equals, uint64(1234))
	c.Assert(e.work.executable, Equals, true)
	c.Assert(e.work.packedHash, Equals, "abcd124befbde41392409")
	c.Assert(e.basis.minikind, Equals, byte('f'))
	c.Assert(e.basis.fingerprint, Equals, "abcd124defbeef")
	c.Assert(e.basis.size, Equals, uint64(9999))
	c.Assert(e.basis.executable, Equals, true)
	c.Assert(e.basis.revisionID, Equals, "rev-id")
	c.Assert(len(e.extra), Equals, 1)
	extra := e.extra[0]
	c.Assert(extra.minikind, Equals, byte('d'))
	c.Assert(extra.size, Equals, uint64(8888))
	c.Assert(extra.fingerprint, Equals, "defbeef")
	c.Assert(extra.executable, Equals, false)
	c.Assert(extra.revisionID, Equals, "alt-rev-id")
}

func (s *WhiteboxSuite) TestEntryParserGetEntryW2Extra(c *C) {
	text := []byte("dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"f\x00abcd124defbeef\x009999\x00y\x00rev-id\x00" +
		"d\x00defbeef\x008888\x00n\x00alt-rev-id\x00" +
		"d\x00defbeef\x008888\x00n\x00alt-rev-id\x00" +
		"\n\x00")
	parser := newEntryParser(text, 3)
	e := entry{}
	err := parser.getEntry(&e)
	c.Assert(err, IsNil)
	c.Assert(e.key.dirname, Equals, "dir")
	c.Assert(e.key.basename, Equals, "name")
	c.Assert(e.key.fileID, Equals, "file_id")
	c.Assert(e.work.minikind, Equals, byte('f'))
	c.Assert(e.work.fingerprint, Equals, "fingerprint")
	c.Assert(e.work.size, Equals, uint64(1234))
	c.Assert(e.work.executable, Equals, true)
	c.Assert(e.work.packedHash, Equals, "abcd124befbde41392409")
	c.Assert(e.basis.minikind, Equals, byte('f'))
	c.Assert(e.basis.fingerprint, Equals, "abcd124defbeef")
	c.Assert(e.basis.size, Equals, uint64(9999))
	c.Assert(e.basis.executable, Equals, true)
	c.Assert(e.basis.revisionID, Equals, "rev-id")
	c.Assert(len(e.extra), Equals, 2)
	extra := e.extra[0]
	c.Assert(extra.minikind, Equals, byte('d'))
	c.Assert(extra.size, Equals, uint64(8888))
	c.Assert(extra.fingerprint, Equals, "defbeef")
	c.Assert(extra.executable, Equals, false)
	c.Assert(extra.revisionID, Equals, "alt-rev-id")
	extra = e.extra[1]
	c.Assert(extra.minikind, Equals, byte('d'))
	c.Assert(extra.size, Equals, uint64(8888))
	c.Assert(extra.fingerprint, Equals, "defbeef")
	c.Assert(extra.executable, Equals, false)
	c.Assert(extra.revisionID, Equals, "alt-rev-id")
}

func assertParserGetEntryFails(c *C, numParents int, text string,
expected_err string) {
	text_b := []byte(text)
	parser := newEntryParser(text_b, numParents)
	e := entry{}
	err := parser.getEntry(&e)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *WhiteboxSuite) TestEntryParserGetEntryFails(c *C) {
	assertParserGetEntryFails(c, 0, "no\x00key",
		"Unable to find basename for key: Could not find trailing null byte")
	assertParserGetEntryFails(c, 0,
		"dir\x00name\x00file_id\x00"+
			"f\x00fingerp",
		"fingerprint missing: Could not find trailing null byte")
	assertParserGetEntryFails(c, 1,
		"dir\x00name\x00file_id\x00"+
			"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00"+
			"f\x00abcd124",
		"failed to read basis: fingerprint missing: Could not find trailing null byte")
	assertParserGetEntryFails(c, 1,
		"dir\x00name\x00file_id\x00"+
			"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00",
		"failed to read basis: minikind missing: Could not find trailing null byte")
	assertParserGetEntryFails(c, 2,
		"dir\x00name\x00file_id\x00"+
			"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00"+
			"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00",
		"failed to read parent 1: minikind missing: Could not find trailing null byte")
	assertParserGetEntryFails(c, 0,
		"dir\x00name\x00file_id\x00"+
			"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00",
		"missing trailing \\n\\x00")
	// This is a valid 3-parent string, but it should fail to parse at the
	// wrong parent level
	text := "dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"f\x00abcd124defbeef\x009999\x00y\x00rev-id\x00" +
		"d\x00defbeef\x008888\x00n\x00alt-rev-id\x00" +
		"d\x00defbeef\x008888\x00n\x00alt-rev-id\x00" +
		"\n\x00"
	assertParserGetEntryFails(c, 0, text, "missing trailing \\n\\x00")
	assertParserGetEntryFails(c, 1, text, "missing trailing \\n\\x00")
	assertParserGetEntryFails(c, 2, text, "missing trailing \\n\\x00")
}
