package dirstate

import (
	"bufio"
	"bytes"
	"fmt"
	"hash"
	"hash/crc32"
	"io"
	"os"
	"strconv"
)

type ParseError string

func (p ParseError) String() string {
	return string(p)
}

type ReadSlicer interface {
	ReadSlice(delim byte) (line []byte, err os.Error)
}

func readPrelude(r ReadSlicer) (info headerInfo, err os.Error) {
	headerLine, err := r.ReadSlice('\n')
	if err != nil {
		return
	}
	if !bytes.Equal(headerLine, []byte(HEADER_FORMAT_3)) {
		err = ParseError(fmt.Sprintf("Invalid header\n"+
			"should be: %s"+
			"    found: %s",
			HEADER_FORMAT_3, headerLine))
		return
	}
	line, err := r.ReadSlice('\n')
	if err != nil {
		err = ParseError(fmt.Sprintf("Unable to read crc32 line: %s",
			err))
		return
	}
	if !bytes.Equal(line[0:7], []byte("crc32: ")) {
		err = ParseError(fmt.Sprintf("Invalid 'crc32' line. Found: %s", line))
		return
	}
	// I miss line[5:-1] syntax...
	info.Crc32, err = strconv.Atoui(string(line[7 : len(line)-1]))
	if err != nil {
		err = ParseError(fmt.Sprintf("Invalid crc32: %s", err))
		return
	}
	line, err = r.ReadSlice('\n')
	if err != nil {
		err = ParseError(fmt.Sprintf("Unable to read num_entries line: %s",
			err))
		return
	}
	if !bytes.Equal(line[0:13], []byte("num_entries: ")) {
		err = ParseError(fmt.Sprintf("Invalid 'num_entries' line. Found: %s",
			line))
		return
	}
	info.NumEntries, err = strconv.Atoui(string(line[13 : len(line)-1]))
	if err != nil {
		err = ParseError(fmt.Sprintf("Invalid num_entries: %s", err))
	}
	return
}

type HashReader struct {
	h   hash.Hash32
	buf *bufio.Reader
}

func (h *HashReader) ReadSlice(delim byte) (line []byte, err os.Error) {
	line, err = h.buf.ReadSlice(delim)
	if err != nil {
		return
	}
	_, err = h.h.Write(line)
	return
}

// Ideally this would be a hidden func, but it makes testing harder, so I'm
// making it exported
func ParseLenDelimitedLine(line []byte) (result []string, err os.Error) {
	if len(line) < 3 { // must have a value and a newline
		err = ParseError("Invalid len-delimited line: too short")
		return
	}
	chunks := bytes.Split(line[:len(line)-2], []byte{0}, -1)
	num, err := strconv.Atoi(string(chunks[0]))
	if err != nil {
		err = ParseError(fmt.Sprintf("Invalid len-delimited line: %s", err))
		return
	}
	result = make([]string, 0, num)
	for _, x := range chunks[1:] {
		result = append(result, string(x))
	}
	if num != len(result) {
		err = ParseError(fmt.Sprintf(
			"Invalid len-delimited line: Expected %d entries, found %d",
			num, len(result)))
		result = nil
	}
	return
}

func readParentsGhosts(r ReadSlicer) (parents, ghosts []string, err os.Error) {
	line, err := r.ReadSlice('\n')
	if err != nil {
		err = ParseError(fmt.Sprintf("Unable to read parents line: %s", err))
		return
	}
	parents, err = ParseLenDelimitedLine(line)
	if err != nil {
		return
	}
	r.ReadSlice('\x00')
	line, err = r.ReadSlice('\n')
	if err != nil {
		err = ParseError(fmt.Sprintf("Unable to read ghosts line: %s", err))
		return
	}
	r.ReadSlice('\x00')
	ghosts, err = ParseLenDelimitedLine(line)
	return
}

func readHeader(r *bufio.Reader) (state *State, info headerInfo, hr *HashReader, err os.Error) {
	info, err = readPrelude(r)
	if err != nil {
		return
	}
	hr = &HashReader{crc32.NewIEEE(), r}
	parents, ghosts, err := readParentsGhosts(hr)
	state = new(State)
	state.headerState = IN_MEMORY_UNMODIFIED
	state.parents = parents
	state.ghosts = ghosts
	return
}

type entryParser struct {
	text       []byte
	numParents int
	curDirname	[]byte
	curDirnameS	string
}

func newEntryParser(text []byte, numParents int) (e *entryParser) {
	return &entryParser{text, numParents, make([]byte, 0), ""}
}

// Similar to the C 'memchr' function, but with built-in bounds checking
// bytes.Index() is similar, but apparently *much* slower because it is
// actually memmem (searching for a multi-bytes match)
func memchr(text []byte, chr byte) int {
	var end int
	var last int
	last = len(text)
	for end = 0; end < last; end++ {
		if text[end] == chr {
			return end
		}
	}
	return -1
}

// Extract the next byte string from text
func (p *entryParser) getNext() (val []byte, err os.Error) {
	// Index is shown as the #1 or #2 time spent, so inline it/specialize it.
	// compare to end := bytes.Index(p.text, []byte{'\x00'})
	// When parsing a 70k entry dirstate with 2 parents,
	// Total time to parse with bytes.Index 478.092ms
	// Total time to parse with memchr 324.381ms
	// Total time to parse using bytes.IndexByte() 282.541ms
	// It seems that bytes.IndexByte is assembly code, which does all the
	// 'alignment' goodness and SSE instructions to make it extra fast.
	//end := memchr(p.text, '\x00')
	end := bytes.IndexByte(p.text, '\x00')
	if end == -1 {
		// TODO: When issuing an error, we should add the context of what bytes
		//       we are currently processing. Further, Go err returns don't
		//       build up anything like a traceback, so we also probably want
		//       to add that sort of code in. Is there anything like
		//       __file__/__line__ for go?
		//		 seems it is in runtime.Caller(s) and runtime.FuncForPC
		err = ParseError("Could not find trailing null byte")
		return
	}
	val = p.text[:end]
	// I'm hoping this is cheap in go, as we're just moving the slice further
	// along its array
	p.text = p.text[end+1:]
	return
}

func (p *entryParser) getKey(key *entryKey) (err os.Error) {
	// TODO: we probably want to share the dirname strings, so use some state
	//		 to track that.
	// TODO: is there a way to test that objects are the same object, rather
	//		 than just equality?
	var content []byte
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("Unable to find dirname for key: %s", err))
		return
	}
	if bytes.Equal(content, p.curDirname) {
		key.dirname = p.curDirnameS
	} else {
		key.dirname = string(content)
		p.curDirnameS = key.dirname
		p.curDirname = []byte(key.dirname)
	}
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("Unable to find basename for key: %s", err))
		return
	}
	key.basename = string(content)
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("Unable to find fileID for key: %s", err))
		return
	}
	key.fileID = string(content)
	return
}

func (p *entryParser) getDetails(d *details) (err os.Error) {
	var content []byte
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("minikind missing: %s", err))
		return
	}
	if len(content) != 1 {
		err = ParseError("minikind must be a single byte")
		return
	}
	d.minikind = content[0]
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("fingerprint missing: %s", err))
		return
	}
	d.fingerprint = string(content)
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("size missing: %s", err))
		return
	}
	if d.size, err = strconv.Atoui64(string(content)); err != nil {
		err = ParseError(fmt.Sprintf("invalid size: %s", err))
		return
	}
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("executable missing: %s", err))
		return
	}
	if len(content) != 1 {
		err = ParseError("executable must be a single byte long")
		return
	}
	switch content[0] {
	case 'y':
		d.executable = true
	case 'n':
		d.executable = false
	default:
		err = ParseError("executable must be either 'n' or 'y'")
	}
	return
}

func (p *entryParser) getWorkingDetails(d *workingDetails) (err os.Error) {
	var content []byte
	if err = p.getDetails(&d.details); err != nil {
		return
	}
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("packedHash missing: %s", err))
		return
	}
	// TODO: Consider doing a hex decode here to save packedHash as a 20-byte
	//       array, rather than a string
	d.packedHash = string(content)
	return
}

func (p *entryParser) getParentDetails(d *parentDetails) (err os.Error) {
	var content []byte
	if err = p.getDetails(&d.details); err != nil {
		return
	}
	if content, err = p.getNext(); err != nil {
		err = ParseError(fmt.Sprintf("revisionID missing: %s", err))
		return
	}
	d.revisionID = string(content)
	return
}

func (p *entryParser) getEntry(e *entry) (err os.Error) {
	if err = p.getKey(&e.key); err != nil {
		return
	}
	if err = p.getWorkingDetails(&e.work); err != nil {
		return
	}
	if p.numParents > 0 {
		if err = p.getParentDetails(&e.basis); err != nil {
			err = ParseError(fmt.Sprintf("failed to read basis: %s", err))
			return
		}
	}
	for i := 1; i < p.numParents; i++ {
		parent := parentDetails{}
		if err = p.getParentDetails(&parent); err != nil {
			err = ParseError(fmt.Sprintf("failed to read parent %d: %s", i, err))
			return
		}
		e.extra = append(e.extra, parent)
	}
	if len(p.text) < 1 || p.text[0] != '\n' {
		err = ParseError("missing trailing \\n\\x00")
		return
	}
	return
}

func readEntries(state *State, r *bufio.Reader, numEntries uint) (err os.Error) {
	p := newEntryParser(make([]byte, 0), len(state.parents) - len(state.ghosts))
	entries := make([]entry, numEntries)
	for _, e := range entries {
		// The content is actually delimited with '\n\0'
		if p.text, err = r.ReadSlice('\n'); err != nil {
			return
		}
		if err = p.getEntry(&e); err != nil {
			return
		}
		// Consume the next '\0'
		tail, err := r.ReadByte()
		if err != nil {
			return
		}
		if tail != '\x00' {
			err = ParseError("entry was missing final '\\x00'")
		}
	}
	state.entries = entries
	return
}

func Parse(r io.Reader) (state *State, err os.Error) {
	bufReader := bufio.NewReader(r)
	state, info, _, err := readHeader(bufReader)
	if err != nil {
		return
	}
	readEntries(state, bufReader, info.NumEntries)
	return
}
