package dirstate_test

import (
	"bytes"
	"dirstate"
	. "launchpad.net/gocheck"
)

type ParserSuite struct{}

var _ = Suite(&ParserSuite{})


func (s *ParserSuite) TestParseEmptyHeader(c *C) {
	buf := bytes.NewBufferString(`#bazaar dirstate flat format 3
crc32: 123456
num_entries: 0
` +
		"0\x00\n\x00" +
		"0\x00\n\x00")
	state, err := dirstate.Parse(buf)
	c.Assert(err, IsNil)
	c.Assert(state, NotNil)
	c.Assert(state.Parents(), Equals, make([]string, 0))
	c.Assert(state.Ghosts(), Equals, make([]string, 0))
}

func (s *ParserSuite) TestParseSingleEntry(c *C) {
	buf := bytes.NewBufferString(`#bazaar dirstate flat format 3
crc32: 123456
num_entries: 1
` +
		"0\x00\n\x00" +
		"0\x00\n\x00" +
		"dir\x00name\x00file_id\x00" +
		"f\x00fingerprint\x001234\x00y\x00abcd124befbde41392409\x00" +
		"\n\x00")
	state, err := dirstate.Parse(buf)
	c.Assert(err, IsNil)
	c.Assert(state, NotNil)
	c.Assert(state.Parents(), Equals, make([]string, 0))
	c.Assert(state.Ghosts(), Equals, make([]string, 0))
	c.Assert(state.NumEntries(), Equals, 1)
}

func (s *ParserSuite) TestParseParentsAndGhosts(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 123456\n" +
		"num_entries: 0\n" +
		"2\x00parent1\x00parent2\x00\n\x00" +
		"1\x00parent2\x00\n\x00")
	state, err := dirstate.Parse(buf)
	c.Assert(err, IsNil)
	c.Assert(state.Parents(), Equals, [...]string{"parent1", "parent2"}[:])
	c.Assert(state.Ghosts(), Equals, [...]string{"parent2"}[:])
}

func (s *ParserSuite) TestParseInvalidHeader(c *C) {
	buf := bytes.NewBufferString("this is not\n" +
		"a dirstate file\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, "Invalid header\n"+
		"should be: "+dirstate.HEADER_FORMAT_3+
		"    found: this is not\n")
}

func (s *ParserSuite) TestParseEOFAtCRC(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3)
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Unable to read crc32 line: EOF")
}

func (s *ParserSuite) TestParseInvalidCRC(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"not a crc line\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid 'crc32' line. Found: not a crc line\n")
}

func (s *ParserSuite) TestParseNaNCRC(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: abcd\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid crc32: parsing \"abcd\": invalid argument")
}

func (s *ParserSuite) TestParseEOFAtNumEntries(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Unable to read num_entries line: EOF")
}

func (s *ParserSuite) TestParseInvalidNumEntries(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"not num_entries\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid 'num_entries' line. Found: not num_entries\n")
}

func (s *ParserSuite) TestParseNaNNumEntries(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"num_entries: abcd\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid num_entries: parsing \"abcd\": invalid argument")
}

func (s *ParserSuite) TestParseEOFAtParents(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"num_entries: 1\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, "Unable to read parents line: EOF")
}

func (s *ParserSuite) TestParseInvalidParents(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"num_entries: 1\n" +
		"not parents\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid len-delimited line: parsing \"not parent\": invalid argument")
}

func (s *ParserSuite) TestParseEOFAtGhosts(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"num_entries: 1\n" +
		"0\x00\n\x00")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, "Unable to read ghosts line: EOF")
}

func (s *ParserSuite) TestParseInvalidGhosts(c *C) {
	buf := bytes.NewBufferString(dirstate.HEADER_FORMAT_3 +
		"crc32: 12345678\n" +
		"num_entries: 1\n" +
		"0\x00\n\x00" +
		"not ghosts\n")
	_, err := dirstate.Parse(buf)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals,
		"Invalid len-delimited line: parsing \"not ghost\": invalid argument")
}

func assertDelimited(c *C, line string, expected []string) {
	result, err := dirstate.ParseLenDelimitedLine([]byte(line))
	c.Assert(err, IsNil)
	c.Assert(result, Equals, expected)
}

func assertFailDelimited(c *C, line string, expected_err string) {
	result, err := dirstate.ParseLenDelimitedLine([]byte(line))
	c.Assert(result, IsNil)
	c.Assert(err, NotNil)
	c.Assert(err.String(), Equals, expected_err)
}

func (s *ParserSuite) TestParseLenDelimitedLine(c *C) {
	assertDelimited(c, "0\x00\n", make([]string, 0))
	assertDelimited(c, "1\x00hello\x00\n", [...]string{"hello"}[:])
	assertDelimited(c, "2\x00hello\x00there\x00\n",
		[...]string{"hello", "there"}[:])
}

func (s *ParserSuite) TestParseInvalidLenDelimitedLine(c *C) {
	assertFailDelimited(c, "",
		"Invalid len-delimited line: too short")
	assertFailDelimited(c, "\n",
		"Invalid len-delimited line: too short")
	assertFailDelimited(c, "\x00\n",
		"Invalid len-delimited line: too short")
	assertFailDelimited(c, "1\n",
		"Invalid len-delimited line: too short")
	assertFailDelimited(c, "1\x00\n",
		"Invalid len-delimited line: Expected 1 entries, found 0")
	assertFailDelimited(c, "0\x00test\x00\n",
		"Invalid len-delimited line: Expected 0 entries, found 1")
	assertFailDelimited(c, "2\x00test\x00\n",
		"Invalid len-delimited line: Expected 2 entries, found 1")
	assertFailDelimited(c, "abcd\n",
		"Invalid len-delimited line: parsing \"abc\": invalid argument")
}
